﻿Option Explicit On
Option Strict On

Imports iTextSharp
Imports iTextSharp.text
Imports iTextSharp.text.pdf

'
' Crée par SharpDevelop.
' Utilisateur: Sebastien Leroux
' Date: 06/06/2011
' Heure: 15:20
'

Namespace DaaSeb.PdfReport.Charts

Public Structure Point
		Dim x As Single
		dim y as Single
End Structure
	
#Region "PARTIE GRAPHIQUE"

Public Class PieChart
	
	Private _elements As List(Of PieChartElement)
	Private _colorList As List(Of BaseColor)
	Private _width As Single
	Private _height As Single
	Private _rPie As Single
	Private _legend As Legend
	Private _baseFont As BaseFont
	Private _font As Font
	Private _fontSize As Single
	
	Public ShadowOffset As Single
	Public SeparationColor As BaseColor
	Public SeparationWidth As Single
	Public IncludeBorder As Boolean
	Public NegativeWeightColor as BaseColor
	
	Public Sub New(ByVal chartWidth As Single, ByVal chartHeight As Single, ByVal chartPieRadius As Single)
		_elements = New List(Of PieChartElement)
		_colorList = New List(Of BaseColor)
		
		_width = chartWidth
		_height = chartHeight
		_rPie = chartPieRadius
		_baseFont = BaseFont.CreateFont(BaseFont.HELVETICA, BaseFont.CP1252, BaseFont.NOT_EMBEDDED)
		_fontSize = 8
		_font = New Font(_baseFont, _fontSize)

		'DefaultValues
		ShadowOffset = 1
		SeparationColor = New BaseColor(229, 223, 208)
		SeparationWidth = 0.2
		IncludeBorder = False
		NegativeWeightColor = BaseColor.RED
	End Sub
	
	Public Property Height As Single
		Get
			Return _height
		End Get
		Set(ByVal value As Single)
			_height = value
		End Set
	End Property
	
	Public Property PieRadius As Single
		Get
			Return _rPie
		End Get
		Set(ByVal value As Single)
			_rPie = value
		End Set
	End Property

	Public Property Width As Single
		Get
			Return _width
		End Get
		Set(ByVal value As Single)
			_width = value
		End Set
	End Property

	Public Property BaseFont As BaseFont
		Get
			Return _baseFont
		End Get
		Set(ByVal value As BaseFont)
			_baseFont = value
			_font = New Font(_baseFont, _fontSize)
		End Set
	End Property

	Public Property FontSize As Single
		Get
			Return _fontSize
		End Get
		Set(ByVal value As Single)
			_fontSize = value
			_font.Size = value
		End Set
	End Property
	
	Public ReadOnly Property Font As Font
		Get
			return _font
		End Get
	End Property


	'Remove all elements of the pieChart
	Public Sub ClearElements()
		_elements.Clear()
	End Sub
	
	#Region "ColorManagement"
	
	'Add a color to the list
	Public Sub AddColor(ByVal red As Integer, ByVal green As Integer, ByVal blue As Integer)
		_colorList.Add(New BaseColor(red, green, blue))
	End Sub
	
	'Set the complete list of colors
	Public Sub SetColorList(ByRef colorList As List(Of BaseColor))
		_colorList = colorList
	End Sub
		
	'Get a color for an element, from 0 to _nbColors - 1, then descending, then continuing the same cycle
	Private Function GetColor(ByVal index As Integer) As BaseColor
		Dim nbColors, i As Integer
		nbColors = _colorList.Count
		
		If nbColors = 0 Then
			Return BaseColor.BLACK
		End If
		
		i = index Mod (2*nbColors-2)
		
		if i > nbColors-1 then i = 2*nbColors - 2 - i
		
		return _colorList(i)
	End Function

	#End Region
	
	'Construct the legend
	Public Sub ConstructLegend()
		_legend = New Legend(_rPie, _rPie + 10, _width / 2, _height / 2, _width, _height, _font, NegativeWeightColor)
		
		Dim i As Integer
		Dim totWeight As Single
		Dim tmpEl As PieChartElement
		Dim startAngle As Single
		Dim weightAngle As Single
		
		totWeight = 0
		
		For i = 0 To _elements.Count-1
			totWeight += Math.Abs(_elements(i).weight)
		Next i
		
		startAngle = 0
		
		For i = 0 To _elements.Count-1
			tmpEl = _elements(i)
			weightAngle = math.Abs(tmpEl.weight) / totWeight * 360
			_legend.AddElement(GetPercentageString(tmpEl.weight) & " " & tmpEl.name, startAngle - weightAngle/2, (tmpEl.weight<0))
			startAngle -= weightAngle
		Next i
		
		_legend.ConstructLegend()

	End Sub
	
	'Draw the chart at a precise position in the page
	Public Function getTemplate(ByRef cb As PdfContentByte) As PdfTemplate
		Dim tp As PdfTemplate = cb.CreateTemplate(width, height)
		
		Dim i As Integer
		
		Dim totWeight As Single
		totWeight = 0
		For i = 0 To _elements.Count-1
			totWeight += Math.Abs(_elements(i).weight)
		Next i
		
		Dim startAngle As Single
		Dim weightAngle As Single
		dim xCenter, yCenter as Single
		Dim tmpEl As PieChartElement
		
		startAngle = 0
		xCenter = width / 2
		yCenter = height / 2
		
		'Draw the shadow
		Dim shad As PdfShading = PdfShading.SimpleRadial(tp.PdfWriter, CSng(xCenter*1.05), CSng(yCenter*0.95), CSng(_rPie/2), CSng(xCenter*1.05), CSng(yCenter*0.95), CSng(_rPie*1.1), BaseColor.DARK_GRAY, BaseColor.WHITE)
		tp.PaintShading(shad)
		
		'Draw the border of the circle
		shad = PdfShading.SimpleAxial(tp.PdfWriter, CSng(xCenter - _rPie*1.2), 0, CSng(xCenter + _rPie*1.2), 0, BaseColor.WHITE, BaseColor.GRAY)
		Dim pat As PdfShadingPattern = New PdfShadingPattern(shad)
		Dim col As ShadingColor = New ShadingColor(pat)
		tp.SetColorFill(col)
		tp.Circle(xCenter, yCenter, CSng(_rPie*1.04))
		tp.Fill()
				
		'Draw the pie
		For i = 0 To _elements.Count - 1
			tmpEl = _elements(i)
			
			weightAngle = math.Abs(tmpEl.weight) / totWeight * 360
			drawPieArc(tp, xCenter, yCenter, _rPie, startAngle, -weightAngle, GetColor(i))
			startAngle -= weightAngle
			
		Next i
		
		'The legend
		_legend.DrawLegend(tp)
		
		'The rectangle if needed
		If IncludeBorder Then
			tp.SetLineWidth(0.3)
			tp.Rectangle(0, 0, _width, _height)
			tp.Stroke
		End If
		
		Return tp
		
	End Function

	'Get the string for a percentage with on digit, only if it's not 0
	Private Function GetPercentageString(ByVal value As Single) As String	
		Dim v, d, remainder As Double
		
		d = CDbl(0.01)
		v = CDbl(CStr(value))
		
		remainder = v - (d * Math.Round(v/d))
				
		If remainder = 0 Then
			return value.ToString("#0%")
		Else
			return value.ToString("#0.0%")
		End If
	End Function
	
	'Add a new element to the pieChart
	Public Sub AddElement(ByVal name As String, ByVal weight As Single)
		Dim tmpEl As PieChartElement
		tmpEl.name = name
		tmpEl.weight = weight
		
		_elements.Add(tmpEl)
	End Sub
	
	'Reorder the elements : first pass from lower weight to higher weight, and second pass taking the bigger, the smaller, the second bigger...
	Public Sub ReorderElements()
		Dim i, j As Integer
		Dim tmpEl As PieChartElement
		Dim nEl As Integer
		
		nEl = _elements.Count
		For i = 1 To nEl-1
			tmpEl = _elements(i)
			j = i
			
			While j > 0 AndAlso Math.Abs(_elements(j-1).weight) > Math.Abs(tmpEl.weight)
				_elements(j) = _elements(j-1)
				j = j-1
			End While
			
			_elements(j) = tmpEl
		Next i
		
		Dim _elements2 As New List(Of PieChartElement)
		Dim isOdd As Boolean
		Dim n As Integer
		
		n = CInt(Int(nEl/2))
		isOdd = ((nEl Mod 2) <> 0)
		
		For i = 0 To n-1
			_elements2.Add(_elements(nEl-1-i))
			_elements2.Add(_elements(i))
		Next i
		If isOdd Then _elements2.Add(_elements(n))
		
		_elements = _elements2
		
	End Sub
	
	'Special sub to draw pie representing more than 90°, when Bezier curves could be ugly
	Private Sub DrawPieArc(ByRef tp As PdfTemplate, ByVal xCenter As Single, ByVal yCenter As Single, ByVal r As Single _
		, ByVal startAngleDegre As Single, ByVal lengthAngleDegre As Single, ByVal pieColor As BaseColor)
		
		'Get the first and last point on the circle
		Dim m1 As Point
		Dim m2 As Point
		Dim sPi As Single
		
		sPi = CSng(Math.PI)
		
		m1.x = r * CSng(Math.Cos(startAngleDegre * sPi / 180))
		m1.y = r * CSng(Math.Sin(startAngleDegre * sPi / 180))
		
		m2.x = r * CSng(Math.Cos(startAngleDegre * sPi / 180 + lengthAngleDegre * sPi / 180))
		m2.y = r * CSng(Math.Sin(startAngleDegre * sPi / 180 + lengthAngleDegre * sPi / 180))
		
		'Begin to draw
		Dim tmpColor As BaseColor = GetShadowedColor(pieColor, 0.2)
		Dim shad As PdfShading = PdfShading.SimpleAxial(tp.PdfWriter, xCenter + m1.x, yCenter + m1.y, xCenter + m2.x, yCenter + m2.y, tmpColor, pieColor, True, True)
		Dim pat As PdfShadingPattern = New PdfShadingPattern(shad)
		Dim col As ShadingColor = New ShadingColor(pat)
		tp.SetColorFill(col)
		
		tp.MoveTo(xCenter, yCenter)
		tp.LineTo(xCenter + m1.x, yCenter + m1.y)
		
		'Draw the Bézier curves
		Dim nPie As Integer
		nPie = CInt(Math.Ceiling(-lengthAngleDegre / 90))
		
		Dim i As Integer
		Dim tmpStart, tmpLength As Single
		
		tmpStart = startAngleDegre
		
		For i = 1 To nPie
			tmpLength = Math.Max(-90, lengthAngleDegre + (i-1)*90)
			
			DrawPieArc_Sub(tp, xCenter, yCenter, r, tmpStart, tmpLength)
			
			tmpStart -= 90
		Next i
	
		'Close the pie element
		tp.ClosePath()
		tp.Fill()

		'We draw the separation lines
		If separationWidth > 0 Then
			tp.SetColorStroke(separationColor)
			tp.SetLineWidth(separationWidth)
			tp.MoveTo(xCenter + m1.x, yCenter + m1.y)
			tp.LineTo(xCenter, yCenter)
			tp.LineTo(xCenter + m2.x, yCenter + m2.y)
			tp.Stroke()
		End If

	End Sub
		
	'To draw a piece of pie
	Private Sub DrawPieArc_Sub(ByRef tp As PdfTemplate, ByVal xCenter As Single, ByVal yCenter As Single, ByVal r As Single _
		, ByVal startAngleDegre As Single, ByVal lengthAngleDegre As Single)
		
		Dim sPi As Single
		sPi = CSng(Math.PI)
		
		Dim startAngle As Single
		Dim lengthAngle As Single
		
		startAngle = startAngleDegre * sPi / 180
		lengthAngle = lengthAngleDegre * sPi / 180
		
		'Calculation of distance relative to the center of the circle
		
		'Points on the circle
		Dim m1 As Point
		Dim m2 As Point
		
		m1.x = r * CSng(Math.Cos(startAngle))
		m1.y = r * CSng(Math.Sin(startAngle))
		
		m2.x = r * CSng(Math.Cos(startAngle + lengthAngle))
		m2.y = r * CSng(Math.Sin(startAngle + lengthAngle))
		
		'Distance to control points
		Dim dControle As Single
		
		dControle = CSng( r * (4 / 3) * CSng(Math.Tan(lengthAngle/4)) )
		
		'Control points coordinates
		Dim dc1, dc2 As Point
		Dim c1, c2 As Point
		
		dc1.x = dcontrole * CSng(Math.Cos(sPi/2 - startAngle))
		dc1.y = dcontrole * CSng(Math.Sin(sPi/2 - startAngle))
		
		dc2.x = dcontrole * CSng(Math.Cos(sPi/2 - startAngle - lengthAngle))
		dc2.y = dcontrole * CSng(Math.Sin(sPi/2 - startAngle - lengthAngle))
		
		c1.x = m1.x - dc1.x
		c1.y = m1.y + dc1.y
		
		c2.x = m2.x + dc2.x
		c2.y = m2.y - dc2.y
				
		'We draw the curve
		tp.CurveTo(xCenter + c1.x, yCenter + c1.y, xCenter + c2.x, yCenter + c2.y, xCenter + m2.x, yCenter + m2.y)

	End Sub
	
	'Get a darker color
	Public Shared Function GetShadowedColor(ByRef colorInput As BaseColor, ByVal darknessPercentage As Single) As BaseColor
		Dim r, g, b As Integer
		r = colorInput.R
		g = colorInput.G
		b = colorInput.B
		
		r -= CInt( Int( CSng(r) * darknessPercentage ) )
		g -= CInt( Int( CSng(g) * darknessPercentage ) )
		b -= CInt( Int( CSng(b) * darknessPercentage ) )
		
		Return New BaseColor(r, g, b)
		
	End Function
	
	'Get a lighter color
	Public Shared Function GetLighterColor(ByRef colorInput As BaseColor, ByVal darknessPercentage As Single) As BaseColor
		Dim r, g, b As Integer
		r = colorInput.R
		g = colorInput.G
		b = colorInput.B
		
		r += CInt( Int( CSng(255 - r) * darknessPercentage ) )
		g += CInt( Int( CSng(255 - g) * darknessPercentage ) )
		b += CInt( Int( CSng(255 - b) * darknessPercentage ) )
		
		Return New BaseColor(r, g, b)
		
	End Function
	
	'Give a point from its polar coordinates
	Public Shared Function GetPointFromAngle(ByVal r As Single, ByVal angle As Single) As Point
		Dim p As Point
		p.x = r * CSng(Math.Cos(angle * Math.PI/180))
		p.y = r * CSng(Math.Sin(angle * Math.PI/180))
		Return p
	End Function
	
	'To round a single at a given digits number. It helps for some comparing when a 0 is not a real 0 after calculation approximations
	Public Shared Function GetRoundedSingle(ByVal s As Single, ByVal nDigits as Integer) As Single
		return CSng(Math.Round(s*10^nDigits)/10^nDigits)
	End Function
		
	'Element of the PieChart, just a name and a weight
	Private Structure PieChartElement
		Dim name As String
		Dim weight As Single
	End Structure
	
End Class

#End Region
	
#Region "THE LEGEND"

'The global legend
Public Class Legend
	
	Private _rPie As Single 'radius of the pie
	Private _rText As Single 'radius of the legend (to leave some space between
	Private _xCenter As Single 'center point of the pie
	Private _yCenter As Single '    "        "
	Private _width As Single 'total width of the graphic
	private _height as Single 'total height of the graphic
	Private _elements As List(Of LegendElement) 'legend elements
	
	'For the font, it's temporary, it should be integrated in the global project
	Private _font As Font
	
	Private _negativeWeightColor As BaseColor
	
	Public Sub New(ByVal rPie As Single, ByVal rText As Single, ByVal xCenter As Single, ByVal yCenter As Single, ByVal width As Single _
		, ByVal height As Single, ByVal font As Font, ByVal negativeWeightColor As BaseColor)
		
		initializeLegend(rPie, rText, xCenter, yCenter, width, height, font, negativeWeightColor)
		
	End Sub
	
	'Initialization
	Private Sub initializeLegend(ByVal rPie As Single, ByVal rText As Single, ByVal xCenter As Single, ByVal yCenter As Single, ByVal width As Single _
		, ByVal height As Single, ByVal font As Font, ByVal negativeWeightColor As BaseColor)
		
		_rPie = rPie
		_rText = rText
		_xCenter = xCenter
		_yCenter = yCenter
		_width = width
		_height = height
		_elements = New List(Of LegendElement)
		
		_font = font
		
		_negativeWeightColor = negativeWeightColor

	End Sub
	
	'Adding a new legend element
	Public Sub AddElement(ByVal text As String, ByVal angle As Single, ByVal isShort As Boolean)
		_elements.Add(New LegendElement(text, angle, _height, _width, _xCenter, _yCenter, _rText, isShort))
	End Sub
	
	'Draw the legend on the pie chart
	Public Sub DrawLegend(ByRef cb As PdfTemplate)
		
		Dim i As Integer
		Dim p1, p2 As Point
		Dim tmpEl As LegendElement
		cb.SetColorFill(BaseColor.BLACK)
		cb.SetColorStroke(BaseColor.BLACK)
		cb.SetFontAndSize(_font.BaseFont, _font.Size)
		
		For i = 0 To _elements.Count - 1
			tmpEl = _elements(i)
			
			p1 = PieChart.GetPointFromAngle(_rPie, tmpEl.AnglePie)
			p2 = tmpEl.TextPoint
			
			p1.x += _xCenter
			p1.y += _yCenter
						
			'First the line of the legend
			cb.MoveTo(p1.x, p1.y)
			cb.LineTo(p2.x, p2.y)
			cb.LineTo(p2.x + Csng(Iif(tmpEl.isOnTheLeft, -5, 5)), p2.y)
			cb.Stroke()

			'Then the text
			If tmpEl.isShort Then _font.Color = _negativeWeightColor Else _font.Color = BaseColor.BLACK
			
			Dim t As New PdfPTable(1)
			t.TotalWidth = tmpEl.TextWidth
			
			Dim ch As New Chunk(tmpEl.Text, _font)
			dim ph as New Phrase(ch)
			Dim c As New PdfPCell()
			c.PaddingLeft = 2
			c.PaddingRight = 2
			c.VerticalAlignment = Element.ALIGN_MIDDLE
			'Temp : Remove this line to see the box for overlapping
			c.Border = Rectangle.NO_BORDER
			'End Temp
			If tmpEl.IsOnTheLeft Then
				c.HorizontalAlignment = Element.ALIGN_RIGHT
			Else
				c.HorizontalAlignment = Element.ALIGN_LEFT
			End If
			c.Phrase = ph
			t.AddCell(c)

			t.WriteSelectedRows(0, -1, p2.x + Csng(Iif(tmpEl.isOnTheLeft, -5 - tmpEl.textWidth, 5)), tmpEl.yUp, cb)

		Next i
		
	End Sub
	
	'Calcul all the positions
	Public Sub ConstructLegend()
		'First we calculate the elements dimensions
		CalculateAllDimensions(_elements)
		
		'Then we deal with superposing
		'We begin with separate in two parts
		Dim elLeft As New List(Of LegendElement)
		Dim elRight As New List(Of LegendElement)
		
		Dim tmpEl As LegendElement
		
		For i = 1 To _elements.Count
			tmpEl = _elements(i-1)
			if tmpEl.IsOnTheLeft then elLeft.Add(tmpEl) else elRight.Add(tmpEl)
		Next i
		
		'We order by angle
		OrderByAngle(elLeft)
		OrderByAngle(elRight)
		InverseOrder(elLeft) 'To have from the upper to the lower
		
		'While there's an overlapping, we move the legends
		Dim avoidOverflow As Integer
		
		CalculateSpaceBelow(elLeft)
		CalculateSpaceAbove(elLeft)
		avoidOverflow = 0
		While checkOverlap(elLeft) And avoidOverflow < 100
			MoveFirstOverlap(elLeft)

			CalculateAllDimensions(elLeft)
			CalculateSpaceBelow(elLeft)
			CalculateSpaceAbove(elLeft)
			
			avoidOverflow += 1
		End While
		
		CalculateSpaceBelow(elRight)
		CalculateSpaceAbove(elRight)
		avoidOverflow = 0
		While checkOverlap(elRight) And avoidOverflow < 100
			MoveFirstOverlap(elRight)

			CalculateAllDimensions(elRight)
			CalculateSpaceBelow(elRight)
			CalculateSpaceAbove(elRight)

			avoidOverflow += 1
		End While
		
	End Sub
	
	'Find the first item which overlap and move it
	Private Sub MoveFirstOverlap(ByRef els As List(Of LegendElement))
		Dim i As Integer
		Dim tmpEl As LegendElement
		
		For i = 0 To els.Count - 1
			tmpEl = els(i)
			
			If tmpEl.IsOverlappingUp or tmpEl.IsOverlappingDown Then
				
				MoveOverlap(els, i)
				Exit Sub
				
			End If
		Next i
		
	End Sub
	
	'Move and item, depending if it overlaps up, down or both
	Private Sub MoveOverlap(ByRef els As List(Of LegendElement), ByVal index As Integer)
		Dim spaceUp, spaceDown, overlapHeight As Single
		dim moveUp, moveDown as Single
		Dim el, tmpEl As LegendElement
		
		el = els(index)
		spaceUp = el.SpaceAbove
		spaceDown = el.SpaceBelow
		
		If el.IsOverlappingDown And el.IsOverlappingUp Then 'We move up the previous and down the following
			
			Dim ovUp, ovDown As Single
			
			If index = els.Count - 1 Then
				ovDown = -el.yDown
			Else
				tmpEl = els(index + 1)
				ovDown = tmpEl.yUp - el.yDown
			End If
			
			If index = 0 Then
				ovUp = el.yUp - _height
			Else
				tmpEl = els(index - 1)
				ovUp = el.yUp - tmpEl.yDown
			End If
			
			overlapHeight = ovUp + OvDown

			If overlapHeight > spaceUp + spaceDown Then overlapHeight = spaceUp + SpaceDown
			
			moveUp = overlapHeight * spaceUp / (spaceUp + spaceDown)
			moveDown = overlapHeight * spaceDown / (spaceUp + spaceDown)
			
			if index > 0 then MoveOverlapUp(els, index - 1, moveUp)
			if index < els.Count - 1 then MoveOverlapDown(els, index + 1, moveDown)
			
		ElseIf el.IsOverlappingDown Then 'We move up this one and down the following
			
			If index = els.Count - 1 Then
				overlapHeight = -el.yDown
			Else
				tmpEl = els(index + 1)
				overlapHeight = tmpEl.yUp - el.yDown
			End If
			
			If overlapHeight > spaceUp + spaceDown Then overlapHeight = spaceUp + SpaceDown
			
			moveUp = overlapHeight * spaceUp / (spaceUp + spaceDown)
			moveDown = overlapHeight * spaceDown / (spaceUp + spaceDown)
			
			MoveOverlapUp(els, index, moveUp)
			if index < els.Count - 1 then MoveOverlapDown(els, index + 1, moveDown)
			
		ElseIf el.IsOverlappingUp Then 'We move down this one and up the previous
			
			If index = 0 Then
				overlapHeight = el.yUp - _height
			Else
				tmpEl = els(index - 1)
				overlapHeight = el.yUp - tmpEl.yDown
			End If
			
			If overlapHeight > spaceUp + spaceDown Then overlapHeight = spaceUp + SpaceDown
			
			moveUp = overlapHeight * spaceUp / (spaceUp + spaceDown)
			moveDown = overlapHeight * spaceDown / (spaceUp + spaceDown)
			
			MoveOverlapDown(els, index, moveDown)
			if index > 0 then MoveOverlapUp(els, index - 1, moveUp)

		End If

	End Sub
	
	'Move up an item, in two passes
	Private Sub MoveOverlapUp(ByRef els As List(Of LegendElement), ByVal index As Integer, ByVal y As Single)
		MoveOverlapUp_FirstPass(els, index, y)
		MoveOverlapUp_SecondPass(els, index, y)
	End Sub
	
	'First pass : move all elements up if space available
	Private Sub MoveOverlapUp_FirstPass(ByRef els As List(Of LegendElement), ByVal index As Integer, ByVal y As Single)
		Dim i As Integer
		Dim tmpMove As Single
		Dim space As Single
		Dim yAbove As Single
		Dim tmpEl As LegendElement
		
		For i = 0 To index
			tmpEl = els(i)
			
			If i = 0 Then yAbove = _height Else yAbove = els(i-1).yDown
			
			space = Math.Max(0, yAbove - tmpEl.yUp)
			
			tmpMove = Math.Min(y, space)
			
			tmpEl.VerticalMove(tmpMove)
		Next i
		
	End Sub

	'Second pass : remove down if space available
	Private Sub MoveOverlapUp_SecondPass(ByRef els As List(Of LegendElement), ByVal index As Integer, ByVal y As Single)
		Dim i As Integer
		Dim tmpMove As single
		Dim space As Single
		Dim yBelow As Single
		Dim tmpEl As LegendElement
		
		For i = index - 1 To 0 Step -1
			tmpEl = els(i)
			
			yBelow = els(i+1).yUp
			
			space = Math.Max(0, tmpEl.yDown - yBelow)
						
			tmpMove = math.Min(y, space)
			
			tmpEl.VerticalMove(-tmpMove)
		Next i
		
	End Sub

	'Move down an item, in two passes
	Private Sub MoveOverlapDown(ByRef els As List(Of LegendElement), ByVal index As Integer, ByVal y As Single)
		MoveOverlapDown_FirstPass(els, index, y)
		MoveOverlapDown_SecondPass(els, index, y)
	End Sub

	'First pass : move all elements down if space available
	Private Sub MoveOverlapDown_FirstPass(ByRef els As List(Of LegendElement), ByVal index As Integer, ByVal y As Single)
		Dim i As Integer
		Dim tmpMove As single
		Dim space As Single
		Dim yBelow As Single
		Dim tmpEl As LegendElement
		
		For i = els.Count - 1 To index Step -1
			tmpEl = els(i)
			
			If i = els.Count-1 Then yBelow = 0 Else yBelow = els(i+1).yUp
			
			space = Math.Max(0, tmpEl.yDown - yBelow)
			
			tmpMove = Math.Min(y, space)
			
			tmpEl.verticalMove(-tmpMove)
		Next i
		
	End Sub
	
	'Second pass : remove up if space available
	Private Sub MoveOverlapDown_SecondPass(ByRef els As List(Of LegendElement), ByVal index As Integer, ByVal y As Single)
		Dim i As Integer
		Dim tmpMove As single
		Dim space As Single
		Dim yAbove As Single
		Dim tmpEl As LegendElement
		
		For i = index + 1 To els.Count - 1
			tmpEl = els(i)
			
			yAbove = els(i-1).yDown
			
			space = Math.Max(0, yAbove - tmpEl.yUp)
			
			tmpMove = math.Min(y, space)
			
			tmpEl.VerticalMove(tmpMove)
		Next i
		
	End Sub
	
	'Call calculate procedure sub for each element
	Private Sub CalculateAllDimensions(ByRef els As List(Of LegendElement))
		Dim i As Integer
				
		For i = 0 To els.Count-1
			els(i).CalculateDimensions(_font)
		Next i
	End Sub
	
	'Calculate total space available below an item
	Private Sub CalculateSpaceBelow(ByRef els As List(Of LegendElement))
		Dim i As Integer
		Dim tmpEl As LegendElement
		
		Dim lastY As Single
		Dim tmpSpace As Single
		lastY = 0
		tmpSpace = 0
		
		For i = els.Count - 1 To 0 Step -1
			tmpEl = els(i)
			
			tmpSpace += Math.Max( 0, tmpEl.yDown - lastY )
			tmpEl.SpaceBelow = tmpSpace
			lastY = tmpEl.yUp
		Next i
		
	End Sub
	
	'Calculate total space available above an item
	Private Sub CalculateSpaceAbove(ByRef els As List(Of LegendElement))
		Dim i As Integer
		Dim tmpEl As LegendElement
		
		Dim lastY As Single
		Dim tmpSpace As Single
		lastY = _height
		tmpSpace = 0
		
		For i = 0 To els.Count - 1
			tmpEl = els(i)
			
			tmpSpace += Math.Max( 0, lastY - tmpEl.yUp )
			tmpEl.SpaceAbove = tmpSpace
			lastY = tmpEl.yDown
		Next i
		
	End Sub

	'Inverse the order of the list of elements
	Private Sub InverseOrder(ByRef els As List(Of LegendElement))
		Dim i As Integer
		Dim els2 As New List(Of LegendElement)
		
		For i = els.Count - 1 To 0 Step -1
			els2.Add(els(i))
		Next i
		
		els = els2
		
	End Sub
	
	'Order a list of elements by angle
	Private Sub OrderByAngle(ByRef els As List(Of LegendElement))
		Dim i, j As Integer
		Dim tmpEl As LegendElement
		Dim nEl As Integer
		
		nEl = els.Count
		For i = 1 To nEl - 1
			tmpEl = els(i)
			j = i
			
			While j > 0 AndAlso ((90 - els(j-1).AnglePie) Mod 360) > ((90 - tmpEl.AnglePie) Mod 360)
				els(j) = els(j-1)
				j = j-1
			End While
			
			els(j) = tmpEl
		Next i
		
	End Sub
	
	'Check if there's an overlap between elements. Return also False if there's one but no more space is available
	Private Function CheckOverlap(ByRef els As List(Of LegendElement)) As Boolean
		Dim i As Integer
		Dim tmpElUp, tmpElDown As LegendElement
		
		Dim oneOverlap As Boolean = False
		
		'Reset the overlapping flag
		For i = 0 To els.Count-1
			els(i).IsOverlappingUp = False
			els(i).IsOverlappingDown = False
		Next i
		
		'Check the first and the last
		tmpElUp = els(0)
		If PieChart.GetRoundedSingle(tmpElUp.yUp, 2) > _height Then tmpElUp.IsOverlappingUp = True : oneOverlap = True
		tmpElDown = els(els.Count - 1)
		If PieChart.GetRoundedSingle(tmpElDown.yDown, 2) < 0 Then tmpElDown.IsOverlappingDown = True : oneOverlap = True
		
		'Check other overlaps
		For i = 0 To els.Count-2
			tmpElUp = els(i)
			tmpElDown = els(i+1)
			
			If PieChart.GetRoundedSingle(tmpElUp.yDown - tmpElDown.yUp, 2) < 0 Then
				tmpElUp.IsOverlappingDown = True
				tmpElDown.IsOverlappingUp = True
				oneOverlap = True
			End If
			
		Next i
		
		'We check if we have space available not to have an infinite process
		Dim totalSpace As Single
		totalSpace = els(0).SpaceAbove + els(0).SpaceBelow
		If PieChart.GetRoundedSingle(totalSpace, 2) <= 0 Then oneOverlap = False

		Return oneOverlap
		
	End Function
	
End Class

'A Legend Element
Public Class LegendElement
	
	Private _text As String 'The text in the legend
	Private _anglePie As Single 'The angle of the middle of the pie part
	Private _textPoint As Point 'The coordinates of the left center point of the legend
	Private _textWidth As Single 'The width of the box
	Private _textHeight As Single 'The height of the box
	Private _isShort As Boolean 'Represents a negative weight
	
	'Properties of the container
	Private _pieHeight As Single
	Private _pieWidth As Single
	Private _pieCenterX As Single
	Private _pieCenterY As Single
	private _pieLegendRadius as Single
	
	'To help placing correctly the legend
	Private _IsOnTheLeft As Boolean
	Public IsOverlappingUp As Boolean
	Public IsOverlappingDown As Boolean
	Public SpaceBelow As Single
	Public SpaceAbove as Single
	
	Public ReadOnly Property Text As String
		Get
			Return _text
		End Get
	End Property

	Public ReadOnly Property AnglePie As Single
		Get
			Return _anglePie
		End Get
	End Property
	
	Public ReadOnly Property TextPoint As Point
		Get
			Return _textPoint
		End Get
	End Property
	
	Public ReadOnly Property TextWidth As Single
		Get
			Return _textWidth
		End Get
	End Property
	
	Public ReadOnly Property TextHeight As Single
		Get
			Return _textHeight
		End Get
	End Property
		
	Public ReadOnly Property IsOnTheLeft As Boolean
		Get
			return _isOnTheLeft
		End Get
	End Property
	
	Public ReadOnly Property yDown As Single
		Get
			Return _textPoint.y - _textHeight/2
		End Get
	End Property

	Public ReadOnly Property yUp As Single
		Get
			Return _textPoint.y + _textHeight/2
		End Get
	End Property
	
	Public ReadOnly Property isShort As Boolean
		Get
			Return _isShort
		End Get
	End Property

	
	Public Sub New(ByVal text As String, ByVal anglePie As Single, ByVal totHeight As Single, ByVal totWidth As Single _
		, ByVal xCenter As Single, ByVal yCenter As Single, ByVal legendRadius As Single)
		
		NewElement(text, anglePie, totHeight, totWidth, xCenter, yCenter, legendRadius, False)

	End Sub

	Public Sub New(ByVal text As String, ByVal anglePie As Single, ByVal totHeight As Single, ByVal totWidth As Single _
		, ByVal xCenter As Single, ByVal yCenter As Single, ByVal legendRadius As Single, ByVal isShort As Boolean)
		
		NewElement(text, anglePie, totHeight, totWidth, xCenter, yCenter, legendRadius, isShort)

	End Sub

	Private Sub NewElement(ByVal text As String, ByVal anglePie As Single, ByVal totHeight As Single, ByVal totWidth As Single _
		, ByVal xCenter As Single, ByVal yCenter As Single, ByVal legendRadius As Single, ByVal isShort As Boolean)
		
		_pieHeight = totHeight
		_pieWidth = totWidth
		_pieCenterX = xCenter
		_pieCenterY = yCenter
		_pieLegendRadius = legendRadius
		_isShort = isShort
		
		_text = text
		_anglePie = anglePie
		_isOnTheLeft = ( _anglePie < -90 And _anglePie >= -270 )
		xyFromAngle()
		
	End Sub
	
	
	'Move vertically the legend, and calculate the horizontal position
	Public Sub verticalMove(ByVal y As Single)
		_textPoint.y += y
		xFromY()
	End Sub
	
	'create the legend box and calculate its dimensions from the horizontal space available
	Public Sub CalculateDimensions(ByRef ft As Font)
		
		'We first calculate the total width available to the left or the right border
		Dim availableWidth As Single
		If isOnTheLeft Then
			availableWidth = _textPoint.x
		Else
			availableWidth = _pieWidth - _textPoint.x
		End If
		availableWidth -= 5 'the small line before the legend
		
		_textWidth = availableWidth
		
		'Then we construct a virtual textArea to calculate the required height
		Dim t As New PdfPTable(1)
		t.TotalWidth = availableWidth
		
		Dim ch As New Chunk(_text, ft)
		Dim ph as New Phrase(ch)
		Dim c As New PdfPCell()
		c.PaddingLeft = 2
		c.PaddingRight = 2
		c.Phrase = ph
		t.AddCell(c)
		
		_textHeight = t.TotalHeight

	End Sub
	
	'get x from y, placing the point on a quadratic curve
	Private Sub xFromY()
		Dim A, B, C As double
		Dim alpha, beta As double
		Dim x, y As Single
		
		A = _pieHeight - _pieCenterY + 5
		B = _pieLegendRadius
		C = _pieCenterY + 5
		
		alpha = - (A*B + B*C) / (A*C^2 + A^2*C)
		beta = (B + C^2*alpha) / C
		
		y = _textPoint.y - _pieCenterY
		x = CSng( alpha * y^2 + beta * y + B )
		
		If Me.isOnTheLeft Then
			_textPoint.x = _pieCenterX - x
		Else
			_textPoint.x = _pieCenterX + x
		End If
	
	End Sub
	
	'For initialization, calculate the original coordonates from the pie part angle
	Private Sub xyFromAngle()
		Dim A, B, C As double
		Dim alpha, beta As double
		Dim x, y As Single
		Dim tmpAngle As Double
		Dim isUp As Boolean
		
		tmpAngle = ((_anglePie+720) Mod 360)
		isUp = ( tmpAngle < 180 )
		'If on the left, we pass it on the right for the calculation
		If tmpAngle > 90 And tmpAngle < 270 Then tmpAngle = (540 - tmpAngle) Mod 360
		tmpAngle *= math.PI/180
		
		A = _pieHeight - _pieCenterY + 5
		B = _pieLegendRadius
		C = _pieCenterY + 5
		
		alpha = - (A*B + B*C) / (A*C^2 + A^2*C)
		beta = (B + C^2*alpha) / C
		
		Dim z, delta, y1, y2 As Double
		
		z = beta - 1/math.Tan(tmpAngle)
		
		delta = z^2 - 4*B*alpha
		If delta < 0 Then Throw New ApplicationException("Can't find the coordonate of the legend for such an angle")
		
		y1 = (-z - math.Sqrt(delta)) / (2*alpha)
		y2 = (-z + math.Sqrt(delta)) / (2*alpha)
		
		If isUp Then
			If y1>=0 Then y=CSng(y1)
			If y2>=0 Then y=CSng(y2)
		Else
			If y1<=0 Then y=CSng(y1)
			If y2<=0 Then y=CSng(y2)
		End If
		
		x = CSng( alpha * y^2 + beta * y + B )
		
		_textPoint.y = _pieCenterY + y
		
		If Me.isOnTheLeft Then
			_textPoint.x = _pieCenterX - x
		Else
			_textPoint.x = _pieCenterX + x
		End If

	End Sub
	
End Class

#End Region

End Namespace
